@@ -1,13 +1,17 @@
 import logging
 import re
 
+from difflib import Differ
 from inspect import Parameter
 from typing import List, Dict, Optional, Any, Iterator, Callable, Union
 
 from jedi import (Script, create_environment,  # type: ignore
                   get_default_environment,
-                  settings as jedi_settings, get_default_project)
+                  settings as jedi_settings, get_default_project,
+                  RefactoringError)
 from jedi.api.classes import Name, Completion  # type: ignore
+from jedi.api.refactoring import Refactoring, ChangedFile  # type: ignore
+from parso import split_lines  # type: ignore
 
 from pycodestyle import (BaseReport as CodestyleBaseReport,  # type: ignore
                          Checker as CodestyleChecker,
@@ -20,7 +24,7 @@ from pygls.features import (COMPLETION, TEXT_DOCUMENT_DID_CHANGE,
                             HOVER, SIGNATURE_HELP, DEFINITION,
                             REFERENCES, WORKSPACE_DID_CHANGE_CONFIGURATION,
                             TEXT_DOCUMENT_WILL_SAVE, TEXT_DOCUMENT_DID_SAVE,
-                            DOCUMENT_SYMBOL)
+                            DOCUMENT_SYMBOL, CODE_ACTION)
 from pygls import types
 from pygls.server import LanguageServer
 from pygls.protocol import LanguageServerProtocol
@@ -98,6 +102,10 @@ class AnakinLanguageServerProtocol(LanguageServerProtocol):
             change=types.TextDocumentSyncKind.INCREMENTAL,
             save=types.SaveOptions()
         )
+        result.capabilities.codeActionProvider = types.CodeActionOptions([
+            types.CodeActionKind.RefactorInline,
+            types.CodeActionKind.RefactorExtract
+        ])
         return result
 
 
@@ -119,6 +127,8 @@ config = {
     'mypy_enabled': False
 }
 
+differ = Differ()
+
 
 def get_script(ls: LanguageServer, uri: str, update: bool = False) -> Script:
     result = None if update else scripts.get(uri)
@@ -645,3 +655,90 @@ def document_symbol(
         script.get_names(all_scopes=True)
     )
     return result
+
+
+def _get_text_edits(changes: ChangedFile) -> List[types.TextEdit]:
+    result = []
+    old_lines = split_lines(changes._module_node.get_code(), keepends=True)
+    new_lines = split_lines(changes.get_new_code(), keepends=True)
+    line_number = 0
+    start = None
+    replace_lines = False
+    lines: List[str] = []
+
+    def _append():
+        if replace_lines:
+            end = types.Position(line_number)
+        else:
+            end = start
+        result.append(
+            types.TextEdit(
+                types.Range(start, end),
+                ''.join(lines)
+            )
+        )
+
+    for line in differ.compare(old_lines, new_lines):
+        kind = line[0]
+        if kind == '?':
+            continue
+        if kind == '-':
+            if not start:
+                start = types.Position(line_number)
+            replace_lines = True
+            line_number += 1
+            continue
+        if kind == '+':
+            if not start:
+                start = types.Position(line_number)
+            lines.append(line[2:])
+            continue
+        if start:
+            _append()
+            start = None
+            replace_lines = False
+            lines = []
+        line_number += 1
+    if start:
+        _append()
+    return result
+
+
+def _get_document_changes(
+        ls: LanguageServer, refactoring: Refactoring
+) -> List[types.TextDocumentEdit]:
+    result = []
+    for fn, changes in refactoring.get_changed_files().items():
+        text_edits = _get_text_edits(changes)
+        if text_edits:
+            uri = from_fs_path(fn)
+            result.append(types.TextDocumentEdit(
+                types.VersionedTextDocumentIdentifier(
+                    uri,
+                    ls.workspace.get_document(uri).version
+                ),
+                text_edits
+            ))
+    return result
+
+
+@server.feature(CODE_ACTION)
+def code_action(
+        ls: LanguageServer, params: types.CodeActionParams
+) -> Optional[List[types.CodeAction]]:
+    if params.range.start != params.range.end:
+        # No selection actions
+        return None
+    script = get_script(ls, params.textDocument.uri)
+    try:
+        refactoring = script.inline(params.range.start.line + 1,
+                                    params.range.start.character)
+    except RefactoringError:
+        return None
+    document_changes = _get_document_changes(ls, refactoring)
+    if document_changes:
+        return [types.CodeAction(
+            'Inline variable',
+            types.CodeActionKind.RefactorInline,
+            edit=types.WorkspaceEdit(document_changes=document_changes))]
+    return None
